@@ -11,10 +11,9 @@ import yaml
 
 from mir.commands import base
 from mir.protos import mir_command_pb2 as mirpb
-from mir.tools import checker, data_exporter, hash_utils, mir_storage_ops, revs_parser
+from mir.tools import checker, class_ids, data_exporter, hash_utils, mir_storage_ops, revs_parser
 from mir.tools.code import MirCode
-
-from ymir.ids import class_ids
+from mir.tools.phase_logger import phase_logger_in_out
 
 
 def _process_model_storage(out_root: str, config_file_path: str, model_upload_location: str,
@@ -118,12 +117,7 @@ def _update_mir_tasks(mir_root: str, base_branch: str, dst_branch: str, task_id:
 # add this function for mock unit test.
 def _run_train_cmd(cmd: str) -> int:
     logging.info("training with cmd: {}".format(cmd))
-    run_result = subprocess.run(cmd.split(" "))  # run and wait
-
-    # when train process done
-    if run_result.returncode != 0:
-        logging.error("training error occured: {}".format(run_result.returncode))
-        return MirCode.RC_CMD_ERROR_UNKNOWN
+    subprocess.run(cmd.split(" "), check=True)  # run and wait, if non-zero value returned, raise
 
     return MirCode.RC_OK
 
@@ -141,19 +135,22 @@ class CmdTrain(base.BaseCommand):
     def run(self) -> int:
         logging.debug("command train: %s", self.args)
 
-        return CmdTrain.run_with_args(export_root=self.args.work_dir,
+        return CmdTrain.run_with_args(work_dir=self.args.work_dir,
                                       model_upload_location=self.args.model_path,
                                       src_revs=self.args.src_revs,
                                       dst_rev=self.args.dst_rev,
                                       mir_root=self.args.mir_root,
                                       media_location=self.args.media_location,
                                       executor=self.args.executor,
+                                      executor_instance=self.args.executor_instance,
                                       config_file=self.args.config_file)
 
     @staticmethod
-    def run_with_args(export_root: str,
+    @phase_logger_in_out
+    def run_with_args(work_dir: str,
                       model_upload_location: str,
                       executor: str,
+                      executor_instance: str,
                       src_revs: str,
                       dst_rev: str,
                       config_file: Optional[str],
@@ -172,8 +169,8 @@ class CmdTrain(base.BaseCommand):
         dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev)
         if checker.check_dst_rev(dst_typ_rev_tid) != MirCode.RC_OK:
             return MirCode.RC_CMD_INVALID_ARGS
-        if not export_root:
-            logging.error("empty export_root, abort")
+        if not work_dir:
+            logging.error("empty work_dir, abort")
             return MirCode.RC_CMD_INVALID_ARGS
         if not config_file:
             logging.warning('empty --config-file, abort')
@@ -196,11 +193,14 @@ class CmdTrain(base.BaseCommand):
             logging.error(f"dumplicate class names in class_names: {class_names}, abort")
             return MirCode.RC_CMD_INVALID_ARGS
 
-        return_code = checker.check(mir_root, [checker.Prerequisites.IS_INSIDE_MIR_REPO])
+        return_code = checker.check(mir_root,
+                                    [checker.Prerequisites.IS_INSIDE_MIR_REPO, checker.Prerequisites.HAVE_LABELS])
         if return_code != MirCode.RC_OK:
             return return_code
 
         task_id = dst_typ_rev_tid.tid
+        if not executor_instance:
+            executor_instance = f"default-training-{task_id}"
 
         # get train_ids, val_ids, test_ids
         train_ids = set()  # type: Set[str]
@@ -235,15 +235,15 @@ class CmdTrain(base.BaseCommand):
 
         # export
         logging.info("exporting assets")
-        os.makedirs(export_root, exist_ok=True)
-        work_dir_in = os.path.join(export_root, "in")
-        work_dir_out = os.path.join(export_root, "out")
+        os.makedirs(work_dir, exist_ok=True)
+        work_dir_in = os.path.join(work_dir, "in")
+        work_dir_out = os.path.join(work_dir, "out")
         os.makedirs(work_dir_in, exist_ok=True)
         os.makedirs(work_dir_out, exist_ok=True)
 
         # type names to type ids
         # ['cat', 'person'] -> [4, 2]
-        cls_mgr = class_ids.ClassIdManager()
+        cls_mgr = class_ids.ClassIdManager(mir_root=mir_root)
         type_ids_list = cls_mgr.id_for_names(class_names)
         if not type_ids_list:
             logging.info(f"type ids empty, please check config file: {config_file}")
@@ -315,7 +315,7 @@ class CmdTrain(base.BaseCommand):
         joint_path_binds = " ".join(path_binds)
         shm_size = _get_shm_size(config_file)
         cmd = (f"nvidia-docker run --rm --shm-size={shm_size} {joint_path_binds} --user {os.getuid()}:{os.getgid()} "
-               f"{executor}")
+               f"--name {executor_instance} {executor}")
 
         ret = _run_train_cmd(cmd)
         if ret != MirCode.RC_OK:
@@ -375,6 +375,11 @@ def bind_to_subparsers(subparsers: argparse._SubParsersAction,
                                   dest="executor",
                                   type=str,
                                   help="docker image name for training")
+    train_arg_parser.add_argument('--executor-instance',
+                                  required=False,
+                                  dest='executor_instance',
+                                  type=str,
+                                  help='docker container name for training')
     train_arg_parser.add_argument("--src-revs",
                                   dest="src_revs",
                                   type=str,
